import unittest
from unittest import mock
from unittest.mock import MagicMock

from eth_account import Account
from hexbytes import HexBytes
from ledgerblue.Dongle import Dongle
from ledgereth import SignedTransaction

from gnosis.safe import SafeTx
from gnosis.safe.tests.safe_test_case import SafeTestCaseMixin

from safe_cli.operators.hw_wallets.hw_wallet_manager import (
    HwWalletManager,
    HwWalletType,
    get_hw_wallet_manager,
)
from safe_cli.operators.hw_wallets.ledger_wallet import LedgerWallet


class Testledger_wallet(SafeTestCaseMixin, unittest.TestCase):
    def test_setup_hw_wallet_manager(self):
        # Should support Treezor and Ledger
        hw_wallet_manager = get_hw_wallet_manager()
        self.assertTrue(hw_wallet_manager.is_supported_hw_wallet(HwWalletType.TREZOR))
        self.assertTrue(hw_wallet_manager.is_supported_hw_wallet(HwWalletType.LEDGER))
        self.assertEqual(len(hw_wallet_manager.wallets), 0)

        # Should get the same instance
        other_hw_wallet_manager = get_hw_wallet_manager()
        self.assertEqual(other_hw_wallet_manager, hw_wallet_manager)

    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.init_dongle",
        autospec=True,
        return_value=Dongle(),
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.LedgerWallet.get_address",
        autospec=True,
    )
    def test_get_accounts(
        self, mock_get_address: MagicMock, mock_init_dongle: MagicMock
    ):
        hw_wallet_manager = HwWalletManager()
        addresses = [Account.create().address, Account.create().address]
        derivation_paths = ["44'/60'/0'/0/0", "44'/60'/1'/0/0"]
        mock_get_address.side_effect = addresses
        # Choosing LEDGER because function is mocked for LEDGER
        hw_wallets = hw_wallet_manager.get_accounts(
            HwWalletType.LEDGER, "44'/60'/{i}'/0/0", number_accounts=2
        )
        self.assertEqual(len(hw_wallets), 2)
        for hw_wallet, expected_address, expected_derivation_path in zip(
            hw_wallets, addresses, derivation_paths
        ):
            address, derivation_path = hw_wallet
            self.assertEqual(expected_address, address)
            self.assertEqual(expected_derivation_path, derivation_path)

    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.init_dongle",
        autospec=True,
        return_value=Dongle(),
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.LedgerWallet.get_address",
        autospec=True,
    )
    def test_add_account(
        self, mock_get_address: MagicMock, mock_init_dongle: MagicMock
    ):
        hw_wallet_manager = HwWalletManager()
        derivation_path = "44'/60'/0'/0"
        account_address = Account.create().address
        mock_get_address.return_value = account_address

        self.assertEqual(len(hw_wallet_manager.wallets), 0)
        # Choosing LEDGER because function is mocked for LEDGER
        self.assertEqual(
            hw_wallet_manager.add_account(HwWalletType.LEDGER, derivation_path),
            account_address,
        )

        self.assertEqual(len(hw_wallet_manager.wallets), 1)
        ledger_wallet = list(hw_wallet_manager.wallets)[0]
        self.assertEqual(ledger_wallet.address, account_address)
        self.assertEqual(ledger_wallet.derivation_path, derivation_path)
        # Shouldn't duplicate accounts
        self.assertEqual(
            hw_wallet_manager.add_account(HwWalletType.LEDGER, derivation_path),
            account_address,
        )
        self.assertEqual(len(hw_wallet_manager.wallets), 1)

        # Should accept derivation paths starting with master
        master_derivation_path = "m/44'/60'/0'/0"
        self.assertEqual(
            hw_wallet_manager.add_account(HwWalletType.LEDGER, master_derivation_path),
            account_address,
        )

    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.init_dongle",
        autospec=True,
        return_value=Dongle(),
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.LedgerWallet.get_address",
        autospec=True,
    )
    def test_delete_account(
        self, mock_get_address: MagicMock, mock_init_dongle: MagicMock
    ):
        hw_wallet_manager = HwWalletManager()
        random_address = Account.create().address
        random_address_2 = Account.create().address
        self.assertEqual(len(hw_wallet_manager.wallets), 0)
        self.assertEqual(len(hw_wallet_manager.delete_accounts([random_address])), 0)

        mock_get_address.return_value = random_address_2
        hw_wallet_manager.wallets.add(LedgerWallet("44'/60'/0'/0"))
        self.assertEqual(len(hw_wallet_manager.delete_accounts([random_address])), 0)
        self.assertEqual(len(hw_wallet_manager.wallets), 1)
        self.assertEqual(len(hw_wallet_manager.delete_accounts([])), 0)

        mock_get_address.return_value = random_address
        hw_wallet_manager.wallets.add(LedgerWallet("44'/60'/0'/1"))
        self.assertEqual(len(hw_wallet_manager.wallets), 2)
        self.assertEqual(len(hw_wallet_manager.delete_accounts([random_address])), 1)
        self.assertEqual(len(hw_wallet_manager.wallets), 1)
        hw_wallet_manager.wallets.add(LedgerWallet("44'/60'/0'/1"))
        self.assertEqual(len(hw_wallet_manager.wallets), 2)
        self.assertEqual(
            len(hw_wallet_manager.delete_accounts([random_address, random_address_2])),
            2,
        )
        self.assertEqual(len(hw_wallet_manager.wallets), 0)

    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.create_transaction",
        autospec=True,
        return_value=Dongle(),
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.LedgerWallet.get_address",
        autospec=True,
    )
    @mock.patch(
        "safe_cli.operators.hw_wallets.ledger_wallet.init_dongle",
        autospec=True,
        return_value=Dongle(),
    )
    def test_execute(
        self,
        mock_init_dongle: MagicMock,
        mock_get_address: MagicMock,
        mock_ledger_create_transaction: MagicMock,
    ):
        owner = self.ethereum_test_account
        to = Account.create()
        derivation_path = "44'/60'/0'/0"
        hw_wallet_manager = HwWalletManager()
        mock_get_address.return_value = owner.address
        hw_wallet_manager.add_account(HwWalletType.LEDGER, derivation_path)
        hw_wallet_manager.set_sender(HwWalletType.LEDGER, derivation_path)
        self.assertEqual(hw_wallet_manager.sender.address, owner.address)
        safe = self.deploy_test_safe(
            owners=[owner.address],
            threshold=1,
            initial_funding_wei=self.w3.to_wei(0.1, "ether"),
        )
        safe_tx = SafeTx(
            self.ethereum_client,
            safe.address,
            to.address,
            10,
            b"",
            0,
            200000,
            200000,
            self.gas_price,
            None,
            None,
            safe_nonce=0,
        )
        safe_tx.sign(owner.key)
        tx_parameters = {
            "from": owner.address,
            "gasPrice": safe_tx.w3.eth.gas_price,
            "nonce": 0,
            "gas": safe_tx.recommended_gas(),
        }
        safe_tx.tx = safe_tx.w3_tx.build_transaction(tx_parameters)
        signed_fields = safe_tx.w3.eth.account.sign_transaction(
            safe_tx.tx, private_key=owner.key
        )

        mocked_signed_transaction_response = SignedTransaction(
            nonce=0,
            gas_price=safe_tx.tx["gasPrice"],
            gas_limit=safe_tx.tx["gas"],
            destination=HexBytes(safe_tx.tx["to"]),
            amount=safe_tx.tx["value"],
            data=HexBytes(safe_tx.tx["data"]),
            v=signed_fields.v,
            r=signed_fields.r,
            s=signed_fields.s,
        )

        mock_ledger_create_transaction.return_value = mocked_signed_transaction_response
        tx_hash, tx = hw_wallet_manager.execute_safe_tx(safe_tx)
        self.assertEqual(tx["data"], safe_tx.tx["data"])
        self.assertIsNotNone(
            self.ethereum_client.w3.eth.get_transaction_receipt(tx_hash)
        )



"""
Get the correct addresses for the contracts by testing the deployment addresses using the RPC
Currently using Safe v1.4.1 when available, and 1.3.0 as fallback as they are compatible
https://github.com/gnosis/safe-deployments/tree/main/src/assets/v1.4.1
https://github.com/gnosis/safe-deployments/tree/main/src/assets/v1.3.0
"""

from typing import Sequence

from eth_typing import ChecksumAddress

from gnosis.eth import EthereumClient


def _get_valid_contract(
    ethereum_client: EthereumClient, addresses: Sequence[ChecksumAddress]
) -> ChecksumAddress:
    """
    :param ethereum_client:
    :param addresses:
    :return: First valid contract from the list of addresses provided found in blockchain
    """

    for address in addresses:
        if ethereum_client.is_contract(address):
            return address
    raise ValueError(f"Network {ethereum_client.get_network().name} is not supported")


def get_safe_contract_address(ethereum_client: EthereumClient) -> ChecksumAddress:
    return _get_valid_contract(
        ethereum_client,
        [
            "0x41675C099F32341bf84BFc5382aF534df5C7461a",  # v1.4.1
            "0xd9Db270c1B5E3Bd161E8c8503c55cEABeE709552",  # v1.3.0
            "0x69f4D1788e39c87893C980c06EdF4b7f686e2938",  # v1.3.0
        ],
    )


def get_safe_l2_contract_address(ethereum_client: EthereumClient) -> ChecksumAddress:
    return _get_valid_contract(
        ethereum_client,
        [
            "0x29fcB43b46531BcA003ddC8FCB67FFE91900C762",  # v1.4.1
            "0x3E5c63644E683549055b9Be8653de26E0B4CD36E",  # v1.3.0
            "0xfb1bffC9d739B8D520DaF37dF666da4C687191EA",  # v1.3.0
            "0x1727c2c531cf966f902E5927b98490fDFb3b2b70",  # v1.3.0 zkSync
        ],
    )


def get_default_fallback_handler_address(
    ethereum_client: EthereumClient,
) -> ChecksumAddress:
    return _get_valid_contract(
        ethereum_client,
        [
            "0xfd0732Dc9E303f09fCEf3a7388Ad10A83459Ec99",  # v1.4.1
            "0xf48f2B2d2a534e402487b3ee7C18c33Aec0Fe5e4",  # v1.3.0
            "0x017062a1dE2FE6b99BE3d9d37841FeD19F573804",  # v1.3.0
            "0x2f870a80647BbC554F3a0EBD093f11B4d2a7492A",  # v1.3.0 zkSync
        ],
    )


def get_proxy_factory_address(ethereum_client: EthereumClient) -> ChecksumAddress:
    return _get_valid_contract(
        ethereum_client,
        [
            "0x4e1DCf7AD4e460CfD30791CCC4F9c8a4f820ec67",  # v1.4.1
            "0xa6B71E26C5e0845f74c812102Ca7114b6a896AB2",  # v1.3.0
            "0xC22834581EbC8527d974F8a1c97E1bEA4EF910BC",  # v1.3.0
            "0xDAec33641865E4651fB43181C6DB6f7232Ee91c2",  # v1.3.0 zkSync
        ],
    )


def get_last_multisend_address(ethereum_client: EthereumClient) -> ChecksumAddress:
    return _get_valid_contract(
        ethereum_client,
        [
            "0x38869bf66a61cF6bDB996A6aE40D5853Fd43B526",  # v1.4.1
            "0xA238CBeb142c10Ef7Ad8442C6D1f9E89e07e7761",  # v1.3.0
            "0x998739BFdAAdde7C933B942a68053933098f9EDa",  # v1.3.0
            "0x0dFcccB95225ffB03c6FBB2559B530C2B7C8A912",  # v1.3.0 zkSync
        ],
    )


def get_last_multisend_call_only_address(
    ethereum_client: EthereumClient,
) -> ChecksumAddress:
    return _get_valid_contract(
        ethereum_client,
        [
            "0x40A2aCCbd92BCA938b02010E17A5b8929b49130D",  # v1.4.1
            "0x40A2aCCbd92BCA938b02010E17A5b8929b49130D",  # v1.3.0
            "0xA1dabEF33b3B82c7814B6D82A79e50F4AC44102B",  # v1.3.0
            "0xf220D3b4DFb23C4ade8C88E526C1353AbAcbC38F",  # v1.3.0 zkSync
        ],
    )


def get_last_sign_message_lib_address(
    ethereum_client: EthereumClient,
) -> ChecksumAddress:
    return _get_valid_contract(
        ethereum_client,
        [
            "0xd53cd0aB83D845Ac265BE939c57F53AD838012c9",  # v1.4.1
            "0xA65387F16B013cf2Af4605Ad8aA5ec25a2cbA3a2",  # v1.3.0
            "0x98FFBBF51bb33A056B08ddf711f289936AafF717",  # v1.3.0
            "0x357147caf9C0cCa67DfA0CF5369318d8193c8407",  # v1.3.0 zkSync
        ],
    )



from typing import Optional

from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from ledgerblue.Dongle import Dongle
from ledgereth import create_transaction, sign_typed_data_draft
from ledgereth.accounts import get_account_by_path
from ledgereth.comms import init_dongle
from web3.types import TxParams

from gnosis.safe.signatures import signature_to_bytes

from .hw_wallet import HwWallet
from .ledger_exceptions import raise_ledger_exception_as_hw_wallet_exception


class LedgerWallet(HwWallet):
    @raise_ledger_exception_as_hw_wallet_exception
    def __init__(self, derivation_path: str):
        self.dongle: Optional[Dongle] = None
        self.connect()
        super().__init__(derivation_path)

    @raise_ledger_exception_as_hw_wallet_exception
    def connect(self) -> bool:
        """
        Connect with ledger
        :return: True if connection was successful or False in other case
        """
        self.dongle = init_dongle(self.dongle)

    @raise_ledger_exception_as_hw_wallet_exception
    def get_address(self) -> ChecksumAddress:
        """

        :return: public address for provided derivation_path
        """
        account = get_account_by_path(self.derivation_path)
        return account.address

    @raise_ledger_exception_as_hw_wallet_exception
    def sign_typed_hash(self, domain_hash: bytes, message_hash: bytes) -> bytes:
        """

        :param domain_hash:
        :param message_hash:
        :return: signature bytes
        """
        signed = sign_typed_data_draft(
            domain_hash, message_hash, self.derivation_path, self.dongle
        )

        return signature_to_bytes(signed.v, signed.r, signed.s)

    @raise_ledger_exception_as_hw_wallet_exception
    def get_signed_raw_transaction(
        self, tx_parameters: TxParams, chain_id: int
    ) -> bytes:
        """

        :param chain_id:
        :param tx_parameters:
        :return: raw transaction signed
        """
        signed_transaction = create_transaction(
            destination=tx_parameters["to"],
            amount=tx_parameters["value"],
            gas=tx_parameters["gas"],
            nonce=tx_parameters["nonce"],
            data=tx_parameters.get("data"),
            max_priority_fee_per_gas=tx_parameters.get("maxPriorityFeePerGas"),
            max_fee_per_gas=tx_parameters.get("maxFeePerGas"),
            chain_id=chain_id,
            sender_path=self.derivation_path,
            dongle=self.dongle,
        )
        return HexBytes(signed_transaction.raw_transaction())



import json
from itertools import chain
from typing import Any, Dict, List, Optional, Sequence, Set, Union

from colorama import Fore, Style
from eth_account.messages import defunct_hash_message
from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from prompt_toolkit import HTML, print_formatted_text
from tabulate import tabulate

from gnosis.eth.contracts import get_erc20_contract
from gnosis.eth.eip712 import eip712_encode_hash
from gnosis.safe import SafeOperationEnum, SafeTx
from gnosis.safe.api import SafeAPIException
from gnosis.safe.api.transaction_service_api.transaction_service_messages import (
    get_remove_transaction_message,
)
from gnosis.safe.multi_send import MultiSend, MultiSendOperation, MultiSendTx
from gnosis.safe.safe_signature import SafeSignature, SafeSignatureEOA
from gnosis.safe.signatures import signature_to_bytes

from safe_cli.utils import yes_or_no_question

from . import SafeServiceNotAvailable
from .exceptions import AccountNotLoadedException, NonExistingOwnerException
from .hw_wallets.hw_wallet import HwWallet
from .safe_operator import SafeOperator


class SafeTxServiceOperator(SafeOperator):
    def __init__(self, address: str, node_url: str):
        super().__init__(address, node_url)
        if not self.safe_tx_service:
            raise SafeServiceNotAvailable(
                f"Cannot configure tx service for network {self.network.name}"
            )
        self.require_all_signatures = (
            False  # It doesn't require all signatures to be present to send a tx
        )

    def approve_hash(self, hash_to_approve: HexBytes, sender: str) -> bool:
        raise NotImplementedError("Not supported when using tx service")

    def sign_message(
        self,
        eip191_message: Optional[str] = None,
        eip712_message_path: Optional[str] = None,
    ) -> bool:
        if eip712_message_path:
            try:
                message = json.load(open(eip712_message_path, "r"))
                message_hash = eip712_encode_hash(message)
            except ValueError:
                raise ValueError
        else:
            message = eip191_message
            message_hash = defunct_hash_message(text=message)

        safe_message_hash = self.safe.get_message_hash(message_hash)
        eoa_signers, hw_wallet_signers = self.get_signers()
        safe_signatures: List[SafeSignature] = []
        for eoa_signer in eoa_signers:
            signature_dict = eoa_signer.signHash(safe_message_hash)
            signature = signature_to_bytes(
                signature_dict["v"], signature_dict["r"], signature_dict["s"]
            )
            safe_signatures.append(SafeSignatureEOA(signature, safe_message_hash))

        signatures = SafeSignature.export_signatures(safe_signatures)

        if hw_wallet_signers:
            print_formatted_text(
                HTML(
                    "<ansired>Signing messages is not currently supported by hardware wallets</ansired>"
                )
            )
            return False

        if self.safe_tx_service.post_message(self.address, message, signatures):
            print_formatted_text(
                HTML(
                    "<ansigreen>Message was correctly created on Safe Transaction Service</ansigreen>"
                )
            )
            return True
        else:
            print_formatted_text(
                HTML(
                    "<ansired>Something went wrong creating message on Safe Transaction Service</ansired>"
                )
            )
            return False

    def get_delegates(self):
        delegates = self.safe_tx_service.get_delegates(self.address)
        headers = ["delegate", "delegator", "label"]
        rows = []
        for delegate in delegates:
            row = [delegate["delegate"], delegate["delegator"], delegate["label"]]
            rows.append(row)
        print(tabulate(rows, headers=headers))
        return rows

    def add_delegate(self, delegate_address: str, label: str, signer_address: str):
        signer_account = [
            account for account in self.accounts if account.address == signer_address
        ]
        if not signer_account:
            raise AccountNotLoadedException(signer_address)
        elif signer_address not in self.safe_cli_info.owners:
            raise NonExistingOwnerException(signer_address)
        else:
            signer_account = signer_account[0]
            try:
                self.safe_tx_service.add_delegate(
                    self.address, delegate_address, label, signer_account
                )
                return True
            except SafeAPIException:
                return False

    def remove_delegate(self, delegate_address: str, signer_address: str):
        signer_account = [
            account for account in self.accounts if account.address == signer_address
        ]
        if not signer_account:
            raise AccountNotLoadedException(signer_address)
        elif signer_address not in self.safe_cli_info.owners:
            raise NonExistingOwnerException(signer_address)
        else:
            signer_account = signer_account[0]
            try:
                self.safe_tx_service.remove_delegate(
                    self.address, delegate_address, signer_account
                )
                return True
            except SafeAPIException:
                return False

    def submit_signatures(self, safe_tx_hash: bytes) -> bool:
        """
        Submit signatures to the tx service

        :return:
        """

        safe_tx, tx_hash = self.safe_tx_service.get_safe_transaction(safe_tx_hash)
        safe_tx.signatures = b""  # Don't post again existing signatures
        if tx_hash:
            print_formatted_text(
                HTML(
                    f"<ansired>Tx with safe-tx-hash {safe_tx_hash.hex()} "
                    f"has already been executed on {tx_hash.hex()}</ansired>"
                )
            )
        else:
            owners = self.get_permitted_signers()
            for account in self.accounts:
                if account.address in owners:
                    safe_tx.sign(account.key)
            # Check if there are ledger signers
            if self.hw_wallet_manager.wallets:
                selected_ledger_accounts = []
                for ledger_account in self.hw_wallet_manager.wallets:
                    if ledger_account.address in owners:
                        selected_ledger_accounts.append(ledger_account)
                if len(selected_ledger_accounts) > 0:
                    safe_tx = self.hw_wallet_manager.sign_safe_tx(
                        safe_tx, selected_ledger_accounts
                    )

            if safe_tx.signers:
                self.safe_tx_service.post_signatures(safe_tx_hash, safe_tx.signatures)
                print_formatted_text(
                    HTML(
                        f"<ansigreen>{len(safe_tx.signers)} signatures were submitted to the tx service</ansigreen>"
                    )
                )
                return True
            else:
                print_formatted_text(
                    HTML(
                        "<ansired>Cannot generate signatures as there were no suitable signers</ansired>"
                    )
                )
        return False

    def batch_txs(self, safe_nonce: int, safe_tx_hashes: Sequence[bytes]) -> bool:
        """
        Submit signatures to the tx service. It's recommended to be on Safe v1.3.0 to prevent issues
        with `safeTxGas` and gas estimation.

        :return:
        """

        try:
            multisend = MultiSend(ethereum_client=self.ethereum_client)
        except ValueError:
            print_formatted_text(
                HTML(
                    "<ansired>Multisend contract is not deployed on this network and it's required for "
                    "batching txs</ansired>"
                )
            )

        multisend_txs = []
        for safe_tx_hash in safe_tx_hashes:
            safe_tx, _ = self.safe_tx_service.get_safe_transaction(safe_tx_hash)
            # Check if call is already a Multisend call
            inner_txs = MultiSend.from_transaction_data(safe_tx.data)
            if inner_txs:
                multisend_txs.extend(inner_txs)
            else:
                multisend_txs.append(
                    MultiSendTx(
                        MultiSendOperation.CALL, safe_tx.to, safe_tx.value, safe_tx.data
                    )
                )

        if len(multisend_txs) > 1:
            safe_tx = SafeTx(
                self.ethereum_client,
                self.address,
                multisend.address,
                0,
                multisend.build_tx_data(multisend_txs),
                SafeOperationEnum.DELEGATE_CALL.value,
                0,
                0,
                0,
                None,
                None,
                safe_nonce=safe_nonce,
            )
        else:
            safe_tx.safe_tx_gas = 0
            safe_tx.base_gas = 0
            safe_tx.gas_price = 0
            safe_tx.signatures = b""
            safe_tx.safe_nonce = safe_nonce  # Resend single transaction
        safe_tx = self.sign_transaction(safe_tx)
        if not safe_tx.signatures:
            print_formatted_text(
                HTML("<ansired>At least one owner must be loaded</ansired>")
            )
            return False
        else:
            return self.post_transaction_to_tx_service(safe_tx)

    def execute_tx(self, safe_tx_hash: Sequence[bytes]) -> bool:
        """
        Submit transaction on the tx-service to blockchain

        :return:
        """
        safe_tx, tx_hash = self.safe_tx_service.get_safe_transaction(safe_tx_hash)
        if tx_hash:
            print_formatted_text(
                HTML(
                    f"<ansired>Tx with safe-tx-hash {safe_tx_hash.hex()} "
                    f"has already been executed on {tx_hash.hex()}</ansired>"
                )
            )
        elif len(safe_tx.signers) < self.safe_cli_info.threshold:
            print_formatted_text(
                HTML(
                    f"<ansired>Number of signatures {len(safe_tx.signers)} "
                    f"must reach the threshold {self.safe_cli_info.threshold}</ansired>"
                )
            )
        else:
            return self.execute_safe_transaction(safe_tx)

    def get_balances(self):
        balances = self.safe_tx_service.get_balances(self.address)
        headers = ["name", "balance", "symbol", "decimals", "tokenAddress"]
        rows = []
        for balance in balances:
            if balance["tokenAddress"]:  # Token
                row = [
                    balance["token"]["name"],
                    f"{int(balance['balance']) / 10 ** int(balance['token']['decimals']):.5f}",
                    balance["token"]["symbol"],
                    balance["token"]["decimals"],
                    balance["tokenAddress"],
                ]
            else:  # Ether
                row = [
                    "ETHER",
                    f"{int(balance['balance']) / 10 ** 18:.5f}",
                    "Ξ",
                    18,
                    "",
                ]
            rows.append(row)
        print(tabulate(rows, headers=headers))
        return rows

    def get_transaction_history(self):
        transactions = self.safe_tx_service.get_transactions(self.address)
        headers = ["nonce", "to", "value", "transactionHash", "safeTxHash"]
        rows = []
        last_executed_tx = False
        for transaction in transactions:
            row = [transaction[header] for header in headers]
            data_decoded: Dict[str, Any] = transaction.get("dataDecoded")
            if data_decoded:
                row.append(self.safe_tx_service.data_decoded_to_text(data_decoded))
            if transaction["transactionHash"]:
                if not transaction["isSuccessful"]:
                    # Transaction failed
                    row[0] = Fore.RED + str(row[0])
                else:
                    row[0] = Fore.GREEN + str(
                        row[0]
                    )  # For executed transactions we use green
                    if not last_executed_tx:
                        row[0] = Style.BRIGHT + row[0]
                        last_executed_tx = True
            else:
                row[0] = Fore.YELLOW + str(
                    row[0]
                )  # For non executed transactions we use yellow

            row[0] = Style.RESET_ALL + row[0]  # Reset all just in case
            rows.append(row)

        headers.append("dataDecoded")
        headers[0] = Style.BRIGHT + headers[0]
        print(tabulate(rows, headers=headers))
        return rows

    def prepare_and_execute_safe_transaction(
        self,
        to: str,
        value: int,
        data: bytes,
        operation: SafeOperationEnum = SafeOperationEnum.CALL,
        safe_nonce: Optional[int] = None,
    ) -> bool:
        safe_tx = self.prepare_safe_transaction(
            to, value, data, operation, safe_nonce=safe_nonce
        )
        return self.post_transaction_to_tx_service(safe_tx)

    def post_transaction_to_tx_service(self, safe_tx: SafeTx) -> bool:
        if not yes_or_no_question(
            f"Do you want to send the tx with safe-tx-hash={safe_tx.safe_tx_hash.hex()} to Safe Transaction Service (it will not be executed) "
            + str(safe_tx)
        ):
            return False

        self.safe_tx_service.post_transaction(safe_tx)
        print_formatted_text(
            HTML(
                f"<ansigreen>Tx with safe-tx-hash={safe_tx.safe_tx_hash.hex()} was sent to Safe Transaction service</ansigreen>"
            )
        )
        return True

    def get_permitted_signers(self) -> Set[ChecksumAddress]:
        """
        :return: Owners and delegates, as they also can sign a transaction for the tx service
        """
        owners = super().get_permitted_signers()
        owners.update(
            [
                row["delegate"]
                for row in self.safe_tx_service.get_delegates(self.address)
            ]
        )
        return owners

    # Function that sends all assets to an account (to)
    def drain(self, to: ChecksumAddress):
        balances = self.safe_tx_service.get_balances(self.address)
        safe_txs = []
        safe_tx = None
        for balance in balances:
            amount = int(balance["balance"])
            if balance["tokenAddress"] is None:  # Then is ether
                if amount != 0:
                    safe_tx = self.prepare_safe_transaction(
                        to,
                        amount,
                        b"",
                        SafeOperationEnum.CALL,
                        safe_nonce=None,
                    )
            else:
                transaction = (
                    get_erc20_contract(self.ethereum_client.w3, balance["tokenAddress"])
                    .functions.transfer(to, amount)
                    .build_transaction({"from": self.address, "gas": 0, "gasPrice": 0})
                )
                safe_tx = self.prepare_safe_transaction(
                    balance["tokenAddress"],
                    0,
                    HexBytes(transaction["data"]),
                    SafeOperationEnum.CALL,
                    safe_nonce=None,
                )
            if safe_tx:
                safe_txs.append(safe_tx)
        if len(safe_txs) > 0:
            multisend_tx = self.batch_safe_txs(safe_tx.safe_nonce, safe_txs)
            if multisend_tx is not None:
                self.post_transaction_to_tx_service(multisend_tx)
                print_formatted_text(
                    HTML(
                        "<ansigreen>Transaction to drain account correctly created</ansigreen>"
                    )
                )
        else:
            print_formatted_text(
                HTML("<ansigreen>Safe account is currently empty</ansigreen>")
            )

    def search_account(
        self, address: ChecksumAddress
    ) -> Optional[Union[LocalAccount, HwWallet]]:
        """
        Search the provided address between loaded owners

        :param address:
        :return: LocalAccount or HwWallet of the provided address
        """
        for account in chain(self.accounts, self.hw_wallet_manager.wallets):
            if account.address == address:
                return account

    def remove_proposed_transaction(self, safe_tx_hash: bytes):
        eip712_message = get_remove_transaction_message(
            self.address, safe_tx_hash, self.ethereum_client.get_chain_id()
        )
        message_hash = eip712_encode_hash(eip712_message)
        try:
            safe_tx, _ = self.safe_tx_service.get_safe_transaction(safe_tx_hash)
            signer = self.search_account(safe_tx.proposer)
            if not signer:
                print_formatted_text(
                    HTML(
                        f"<ansired>The proposer with address: {safe_tx.proposer} was not loaded</ansired>"
                    )
                )
                return False

            if isinstance(signer, LocalAccount):
                signature = signer.signHash(message_hash).signature
            else:
                signature = self.hw_wallet_manager.sign_eip712(eip712_message, [signer])

            if len(safe_tx.signers) >= self.safe.retrieve_threshold():
                print_formatted_text(
                    HTML(
                        "<ansired>The transaction has all the required signatures to be executed!!!\n"
                        "This means that the transaction can be executed by a 3rd party monitoring your Safe even after removal!\n"
                        f"Make sure you execute a transaction with nonce {safe_tx.safe_nonce} to void the current transaction"
                        "</ansired>"
                    )
                )

            if not yes_or_no_question(
                f"Do you want to remove the tx with safe-tx-hash={safe_tx.safe_tx_hash.hex()}"
            ):
                return False

            self.safe_tx_service.delete_transaction(safe_tx_hash.hex(), signature.hex())
            print_formatted_text(
                HTML(
                    f"<ansigreen>Transaction {safe_tx_hash.hex()} was removed correctly</ansigreen>"
                )
            )
            return True
        except SafeAPIException as e:
            print_formatted_text(
                HTML(f"<ansired>Transaction wasn't removed due an error: {e}</ansired>")
            )
            return False



import requests
from web3 import Web3
from web3.auto import w3
from eth_account.messages import encode_defunct
import json

class FailSafeAttestationServiceAuth:
    def __init__(self, fs_attestation_service_base_url: str, safe_cli_owner: str, safe_cli_owner_private_key: str):
        self.fs_attestation_service_base_url = fs_attestation_service_base_url
        self.safe_cli_owner = safe_cli_owner
        self.safe_cli_owner_private_key = safe_cli_owner_private_key
        self.fs_api_key = "1Rm0ER9VPn8tNsQRQf7Iia9Koq7I33b728Ml7dvO"
        # TO DO - Remove below line
        self.fs_attestation_service_base_url = "https://1fw43grxlf.execute-api.us-east-1.amazonaws.com/v1"

    def initiate_auth(self):
        endpoint_url = "{}/auth/login".format(self.fs_attestation_service_base_url)
        headers = {
            'x-api-key': self.fs_api_key,
        }
        data = {
            'wallet_address': self.safe_cli_owner.lower(),
        }
        response = requests.post(endpoint_url, json=data, headers=headers)
        return response.json()

    def sign_challenge(self, challenge):
        web3_instance = Web3()
        try:
            message_hash = encode_defunct(text=challenge)
            message_signed = w3.eth.account.sign_message(message_hash, private_key=self.safe_cli_owner_private_key)
            signature_hash = message_signed.signature.hex()
            return signature_hash
        except Exception as error:
            raise

    def verify_challenge(self, signature, session_token):
        endpoint_url = "{}/auth/login".format(self.fs_attestation_service_base_url)
        payload = {
            'wallet_address': self.safe_cli_owner.lower(),
            'challengeResponse': { 'signature': signature },
            'sessionToken': session_token
        }

        # Configuration object with headers including the x-api-key
        headers = {
            'x-api-key': self.fs_api_key
        }

        try:
            response = requests.post(endpoint_url, json=payload, headers=headers)
            response_data = response.json()
            if 'AuthenticationResult' in response_data:
                return response_data
            else:
                raise Exception('Verification failed')
        except Exception as error:
            raise

    # Perform the authentication flow and returns the access token
    def fs_attestation_user_login(self) -> str:
        try:
            initiate_auth_response = self.initiate_auth()
            session_token = initiate_auth_response['Session']
            challenge = initiate_auth_response['ChallengeParameters']['message']

            signature = self.sign_challenge(challenge)
            verification_response = self.verify_challenge(signature, session_token)

            access_token = verification_response['AuthenticationResult']['AccessToken']
            return access_token
        except Exception as error:
            raise


# flake8: noqa F401
from .enums import SafeOperatorMode
from .exceptions import SafeServiceNotAvailable
from .safe_operator import SafeOperator
from .fs_attestation_service_operator import FailSafeAttestationServiceOperator
from .safe_tx_service_operator import SafeTxServiceOperator
from .fs_attestation_auth import FailSafeAttestationServiceAuth


from eth_account import Account
from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress

ETHEREUM_DEFAULT_PATH = "m/44'/60'/0'/0/0"
ETHEREUM_BASE_PATH = "m/44'/60'/0'/0"


def get_account_from_words(
    words: str, index: int = 0, hd_path: str = ETHEREUM_DEFAULT_PATH
) -> LocalAccount:
    """
    :param words: Mnemonic words(BIP39) for a Hierarchical Deterministic Wallet(BIP32)
    :param index: Index of account
    :param hd_path: BIP44 Path. By default Ethereum with 0 index is used
    :return: Ethereum Account
    :raises: eth_utils.ValidationError
    """
    Account.enable_unaudited_hdwallet_features()
    if index:
        hd_path = f"{ETHEREUM_BASE_PATH}/{index}"
    return Account.from_mnemonic(words, account_path=hd_path)


def get_address_from_words(
    words: str, index: int = 0, hd_path: str = ETHEREUM_DEFAULT_PATH
) -> ChecksumAddress:
    """
    :param words: Mnemonic words(BIP39) for a Hierarchical Deterministic Wallet(BIP32)
    :param index: Index of account
    :param hd_path: BIP44 Path. By default Ethereum with 0 index is used
    :return: Ethereum checksummed public address
    :raises: eth_utils.ValidationError
    """
    return get_account_from_words(words, index, hd_path).address



from functools import lru_cache

import rlp
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from trezorlib import tools
from trezorlib.client import TrezorClient, get_default_client
from trezorlib.ethereum import (
    get_address,
    sign_tx,
    sign_tx_eip1559,
    sign_typed_data_hash,
)
from trezorlib.ui import ClickUI
from web3.types import TxParams

from .hw_wallet import HwWallet
from .trezor_exceptions import raise_trezor_exception_as_hw_wallet_exception


@lru_cache(maxsize=None)
@raise_trezor_exception_as_hw_wallet_exception
def get_trezor_client() -> TrezorClient:
    """
    Return default trezor configuration that store passphrase on host.
    This method is cached to share the same configuration between trezor calls while the class is not instantiated.
    :return:
    """
    ui = ClickUI(passphrase_on_host=True, always_prompt=True)
    client = get_default_client(ui=ui)
    return client


class TrezorWallet(HwWallet):
    def __init__(self, derivation_path: str):
        self.client: TrezorClient = get_trezor_client()
        super().__init__(derivation_path)

    @raise_trezor_exception_as_hw_wallet_exception
    def get_address(self) -> ChecksumAddress:
        """
        :return: public address for derivation_path
        """
        address_n = tools.parse_path(self.derivation_path)
        return get_address(client=self.client, n=address_n)

    @raise_trezor_exception_as_hw_wallet_exception
    def sign_typed_hash(self, domain_hash: bytes, message_hash: bytes) -> bytes:
        """

        :param domain_hash:
        :param message_hash:
        :return: signature bytes
        """
        address_n = tools.parse_path(self.derivation_path)
        signed = sign_typed_data_hash(
            self.client, n=address_n, domain_hash=domain_hash, message_hash=message_hash
        )
        return signed.signature

    @raise_trezor_exception_as_hw_wallet_exception
    def get_signed_raw_transaction(
        self, tx_parameters: TxParams, chain_id: int
    ) -> bytes:
        """

        :param chain_id:
        :param tx_parameters:
        :return: raw transaction signed
        """
        address_n = tools.parse_path(self.derivation_path)
        if tx_parameters.get("maxPriorityFeePerGas"):
            # EIP1559
            v, r, s = sign_tx_eip1559(
                self.client,
                n=address_n,
                nonce=tx_parameters["nonce"],
                gas_limit=tx_parameters["gas"],
                to=tx_parameters["to"],
                value=tx_parameters["value"],
                data=HexBytes(tx_parameters["data"]),
                chain_id=chain_id,
                max_gas_fee=tx_parameters.get("maxFeePerGas"),
                max_priority_fee=tx_parameters.get("maxPriorityFeePerGas"),
            )

            encoded_transaction = (
                "0x02"
                + rlp.encode(
                    [
                        chain_id,
                        tx_parameters["nonce"],
                        tx_parameters.get("maxPriorityFeePerGas"),
                        tx_parameters.get("maxFeePerGas"),
                        tx_parameters["gas"],
                        HexBytes(tx_parameters["to"]),
                        tx_parameters["value"],
                        HexBytes(tx_parameters["data"]),
                        [],
                        v,
                        HexBytes(r),
                        HexBytes(s),
                    ]
                ).hex()
            )
        else:
            # Legacy transaction
            v, r, s = sign_tx(
                self.client,
                n=address_n,
                nonce=tx_parameters["nonce"],
                gas_price=tx_parameters["gasPrice"],
                gas_limit=tx_parameters["gas"],
                to=tx_parameters["to"],
                value=tx_parameters["value"],
                data=HexBytes(tx_parameters.get("data")),
                chain_id=chain_id,
            )

            encoded_transaction = rlp.encode(
                [
                    tx_parameters["nonce"],
                    tx_parameters["gasPrice"],
                    tx_parameters["gas"],
                    HexBytes(tx_parameters["to"]),
                    tx_parameters["value"],
                    HexBytes(tx_parameters["data"]),
                    v,
                    HexBytes(r),
                    HexBytes(s),
                ]
            ).hex()

        return HexBytes(encoded_transaction)



import re
from abc import ABC, abstractmethod

from web3.types import TxParams

from .constants import BIP32_ETH_PATTERN, BIP32_LEGACY_LEDGER_PATTERN
from .exceptions import InvalidDerivationPath


class HwWallet(ABC):
    def __init__(self, derivation_path: str):
        derivation_path = derivation_path.replace("m/", "")
        if self._is_valid_derivation_path(derivation_path):
            self.derivation_path = derivation_path
        self.address = self.get_address()

    @property
    def get_derivation_path(self):
        return self.derivation_path

    @abstractmethod
    def get_address(self):
        """

        :return:
        """

    def _is_valid_derivation_path(self, derivation_path: str):
        """
        Detect if a string is a valid derivation path
        """
        if not (
            re.match(BIP32_ETH_PATTERN, derivation_path) is not None
            or re.match(BIP32_LEGACY_LEDGER_PATTERN, derivation_path) is not None
        ):
            raise InvalidDerivationPath()

        return True

    @abstractmethod
    def sign_typed_hash(self, domain_hash: bytes, message_hash: bytes) -> bytes:
        """

        :param domain_hash:
        :param message_hash:
        :return: signature bytes
        """

    @abstractmethod
    def get_signed_raw_transaction(
        self, tx_parameters: TxParams, chain_id: int
    ) -> bytes:
        """

        :param chain_id:
        :param tx_parameters:
        :return: raw transaction signed
        """

    def __str__(self):
        return f"{self.__class__.__name__} device with address {self.address}"

    def __eq__(self, other):
        if isinstance(other, HwWallet):
            return (
                self.derivation_path == other.derivation_path
                and self.address == other.address
            )
        return False

    def __hash__(self):
        return hash((self.derivation_path, self.address))



import argparse
from binascii import Error

from eth_account import Account
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from web3 import Web3


def check_positive_integer(number: str) -> int:
    """
    Positive integer validator for Argparse

    :param number:
    :return: Positive integer
    """
    number = int(number)
    if number <= 0:
        raise argparse.ArgumentTypeError(
            f"{number} is not a valid threshold. Must be > 0"
        )
    return number


def check_ethereum_address(address: str) -> ChecksumAddress:
    """
    Ethereum address validator for Argparse

    :param address:
    :return: Checksummed ethereum address
    """
    if not Web3.is_checksum_address(address):
        raise argparse.ArgumentTypeError(
            f"{address} is not a valid checksummed ethereum address"
        )
    return ChecksumAddress(address)


def check_private_key(private_key: str) -> str:
    """
    Ethereum private key validator for ArgParse

    :param private_key: Ethereum Private key
    :return: Ethereum Private key
    """
    try:
        Account.from_key(private_key)
    except (ValueError, Error):  # TODO Report `Error` exception as a bug of eth_account
        raise argparse.ArgumentTypeError(f"{private_key} is not a valid private key")
    return private_key


def check_hex_str(hex_str: str) -> HexBytes:
    """
    Hexadecimal string validator for Argparse

    :param hex_str:
    :return: HexBytes from the provided hex string
    """
    try:
        return HexBytes(hex_str)
    except ValueError:
        raise argparse.ArgumentTypeError(f"{hex_str} is not a valid hexadecimal string")


def check_keccak256_hash(hex_str: str) -> HexBytes:
    """
    Hexadecimal keccak256 validator for Argparse

    :param hex_str:
    :return: HexBytes from the provided hex string
    """
    hex_str_bytes = check_hex_str(hex_str)
    if len(hex_str_bytes) != 32:
        raise argparse.ArgumentTypeError(
            f"{hex_str} is not a valid keccak256 hash hexadecimal string"
        )
    return hex_str_bytes



import argparse
import functools

from prompt_toolkit import HTML, print_formatted_text

from gnosis.safe.api import SafeAPIException

from .argparse_validators import (
    check_ethereum_address,
    check_hex_str,
    check_keccak256_hash,
)
from .operators import SafeServiceNotAvailable
from .operators.exceptions import (
    AccountNotLoadedException,
    ExistingOwnerException,
    FallbackHandlerNotSupportedException,
    HardwareWalletException,
    HashAlreadyApproved,
    InvalidMasterCopyException,
    InvalidMigrationContractException,
    InvalidNonceException,
    NonExistingOwnerException,
    NotEnoughEtherToSend,
    NotEnoughSignatures,
    NotEnoughTokenToSend,
    SafeAlreadyUpdatedException,
    SafeOperatorException,
    SafeVersionNotSupportedException,
    SameFallbackHandlerException,
    SameMasterCopyException,
    SenderRequiredException,
    ThresholdLimitException,
)
from .operators.safe_operator import SafeOperator
from .operators.fs_attestation_service_operator import FailSafeAttestationServiceOperator

def parse_signers_ip_address_constraints(signers_ip_address_constraints_str: str) -> list:
    """
    Parse the signers IP address constraints
    :param signers_ip_address_constraints:
    :return: Parsed signers IP address constraints
    """
    signers_ip_address_constraints = []
    for constraint in signers_ip_address_constraints_str.split(","):
        signers_ip_address_constraint = {}
        signer, ip_address = constraint.split(":")
        signers_ip_address_constraint["wallet_address"] = signer
        signers_ip_address_constraint["ip_address_range"] = ip_address
        signers_ip_address_constraints.append(signers_ip_address_constraint)
    return signers_ip_address_constraints


def safe_exception(function):
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except SafeAPIException as e:
            if e.args:
                print_formatted_text(HTML(f"<b><ansired>{e.args[0]}</ansired></b>"))
        except AccountNotLoadedException as e:
            print_formatted_text(
                HTML(f"<ansired>Account {e.args[0]} is not loaded</ansired>")
            )
        except NotEnoughSignatures as e:
            print_formatted_text(
                HTML(
                    f"<ansired>Cannot find enough owners to sign. {e.args[0]} missing</ansired>"
                )
            )
        except SenderRequiredException:
            print_formatted_text(
                HTML("<ansired>Please load a default sender</ansired>")
            )
        except ExistingOwnerException as e:
            print_formatted_text(
                HTML(
                    f"<ansired>Owner {e.args[0]} is already an owner of the Safe"
                    f"</ansired>"
                )
            )
        except NonExistingOwnerException as e:
            print_formatted_text(
                HTML(
                    f"<ansired>Owner {e.args[0]} is not an owner of the Safe"
                    f"</ansired>"
                )
            )
        except HashAlreadyApproved as e:
            print_formatted_text(
                HTML(
                    f"<ansired>Transaction with safe-tx-hash {e.args[0].hex()} has already been approved by "
                    f"owner {e.args[1]}</ansired>"
                )
            )
        except ThresholdLimitException:
            print_formatted_text(
                HTML(
                    "<ansired>Having less owners than threshold is not allowed"
                    "</ansired>"
                )
            )
        except SameFallbackHandlerException as e:
            print_formatted_text(
                HTML(
                    f"<ansired>Fallback handler {e.args[0]} is the current one</ansired>"
                )
            )
        except FallbackHandlerNotSupportedException:
            print_formatted_text(
                HTML(
                    "<ansired>Fallback handler is not supported for your Safe, "
                    "you need to <b>update</b> first</ansired>"
                )
            )
        except SameMasterCopyException as e:
            print_formatted_text(
                HTML(f"<ansired>Master Copy {e.args[0]} is the current one</ansired>")
            )
        except InvalidMasterCopyException as e:
            print_formatted_text(
                HTML(f"<ansired>Master Copy {e.args[0]} is not valid</ansired>")
            )
        except InvalidMigrationContractException as e:
            print_formatted_text(HTML(f"<ansired>{e.args[0]}</ansired>"))
        except InvalidNonceException as e:
            print_formatted_text(HTML(f"<ansired>{e.args[0]}</ansired>"))
        except SafeAlreadyUpdatedException:
            print_formatted_text(HTML("<ansired>Safe is already updated</ansired>"))
        except SafeVersionNotSupportedException as e:
            print_formatted_text(HTML(f"<ansired>{e.args[0]}</ansired>"))
        except (NotEnoughEtherToSend, NotEnoughTokenToSend) as e:
            print_formatted_text(
                HTML(
                    f"<ansired>Cannot find enough to send. Current balance is {e.args[0]}"
                    f"</ansired>"
                )
            )
        except SafeServiceNotAvailable as e:
            print_formatted_text(
                HTML(
                    f"<ansired>Service not available for network {e.args[0]}</ansired>"
                )
            )
        except HardwareWalletException as e:
            print_formatted_text(
                HTML(f"<ansired>HwDevice exception: {e.args[0]}</ansired>")
            )
        except SafeOperatorException as e:
            print_formatted_text(HTML(f"<ansired>{e.args[0]}</ansired>"))

    return wrapper


class PromptParser:
    def __init__(self, safe_operator: SafeOperator, fs_attestation_service_operator: FailSafeAttestationServiceOperator):
        self.mode_parser = argparse.ArgumentParser(prog="")
        self.safe_operator = safe_operator
        self.fs_attestation_service_operator = fs_attestation_service_operator
        self.prompt_parser = build_prompt_parser(safe_operator, fs_attestation_service_operator)

    def process_command(self, command: str):
        args = self.prompt_parser.parse_args(command.split())
        return args.func(args)


def build_prompt_parser(safe_operator: SafeOperator, fs_attestation_service_operator: FailSafeAttestationServiceOperator) -> argparse.ArgumentParser:
    """
    Returns an ArgParse capable of decoding and executing the Safe commands
    :param1 safe_operator:
    :param2 fs_attestation_service_operator
    :return:
    """
    prompt_parser = argparse.ArgumentParser(prog="")
    subparsers = prompt_parser.add_subparsers()

    @safe_exception
    def show_cli_owners(args):
        safe_operator.show_cli_owners()

    @safe_exception
    def load_cli_owners_from_words(args):
        safe_operator.load_cli_owners_from_words(args.words)

    @safe_exception
    def load_cli_owners(args):
        safe_operator.load_cli_owners(args.keys)

    @safe_exception
    def load_ledger_cli_owners(args):
        safe_operator.load_ledger_cli_owners(
            derivation_path=args.derivation_path, legacy_account=args.legacy_accounts
        )

    @safe_exception
    def load_trezor_cli_owners(args):
        safe_operator.load_trezor_cli_owners(
            derivation_path=args.derivation_path, legacy_account=args.legacy_accounts
        )

    @safe_exception
    def unload_cli_owners(args):
        safe_operator.unload_cli_owners(args.addresses)

    @safe_exception
    def approve_hash(args):
        safe_operator.approve_hash(args.hash_to_approve, args.sender)

    @safe_exception
    def sign_message(args):
        safe_operator.sign_message(args.eip191_message, args.eip712_path)

    @safe_exception
    def add_owner(args):
        safe_operator.add_owner(args.address, threshold=args.threshold)

    @safe_exception
    def remove_owner(args):
        safe_operator.remove_owner(args.address, threshold=args.threshold)

    @safe_exception
    def change_fallback_handler(args):
        safe_operator.change_fallback_handler(args.address)

    @safe_exception
    def change_guard(args):
        safe_operator.change_guard(args.address)

    @safe_exception
    def change_master_copy(args):
        safe_operator.change_master_copy(args.address)

    @safe_exception
    def change_threshold(args):
        safe_operator.change_threshold(args.threshold)

    @safe_exception
    def send_custom(args):
        safe_operator.send_custom(
            args.to,
            args.value,
            args.data,
            safe_nonce=args.safe_nonce,
            delegate_call=args.delegate,
        )

    @safe_exception
    def send_ether(args):
        safe_operator.send_ether(args.to, args.value, safe_nonce=args.safe_nonce)

    @safe_exception
    def send_erc20(args):
        safe_operator.send_erc20(
            args.to, args.token_address, args.amount, safe_nonce=args.safe_nonce
        )

    @safe_exception
    def send_erc721(args):
        safe_operator.send_erc721(
            args.to, args.token_address, args.token_id, safe_nonce=args.safe_nonce
        )

    @safe_exception
    def drain(args):
        safe_operator.drain(args.to)

    @safe_exception
    def get_threshold(args):
        safe_operator.get_threshold()

    @safe_exception
    def get_nonce(args):
        safe_operator.get_nonce()

    @safe_exception
    def get_owners(args):
        safe_operator.get_owners()

    @safe_exception
    def enable_module(args):
        safe_operator.enable_module(args.address)

    @safe_exception
    def disable_module(args):
        safe_operator.disable_module(args.address)

    @safe_exception
    def update_version(args):
        safe_operator.update_version()

    @safe_exception
    def update_version_to_l2(args):
        safe_operator.update_version_to_l2(args.migration_contract)

    @safe_exception
    def get_info(args):
        safe_operator.print_info()

    @safe_exception
    def get_refresh(args):
        safe_operator.refresh_safe_cli_info()

    @safe_exception
    def get_balances(args):
        safe_operator.get_balances()

    @safe_exception
    def get_history(args):
        safe_operator.get_transaction_history()

    @safe_exception
    def sign_tx(args):
        safe_operator.submit_signatures(args.safe_tx_hash)

    @safe_exception
    def batch_txs(args):
        safe_operator.batch_txs(args.safe_nonce, args.safe_tx_hashes)

    @safe_exception
    def execute_tx(args):
        safe_operator.execute_tx(args.safe_tx_hash)

    @safe_exception
    def get_delegates(args):
        safe_operator.get_delegates()

    @safe_exception
    def add_delegate(args):
        safe_operator.add_delegate(args.address, args.label, args.signer)

    @safe_exception
    def remove_delegate(args):
        safe_operator.remove_delegate(args.address, args.signer)

    @safe_exception
    def remove_proposed_transaction(args):
        safe_operator.remove_proposed_transaction(args.safe_tx_hash)

    # Extension methods for FailSafe Attestation Service
    @safe_exception
    def fs_attestation_service_auth(args):
        fs_attestation_service_operator.fs_attestation_user_login(args.safe_cli_owner, args.safe_cli_owner_private_key)

    @safe_exception
    def fs_attestation_service_enrolment(args):
        fs_attestation_service_operator.fs_attestation_service_enrolment()

    @safe_exception
    def fs_attestation_service_deploy_guard_contract(args):
        fs_attestation_service_operator.fs_attestation_service_deploy_guard_contract()

    @safe_exception
    def fs_attestation_service_set_constraints(args):
        fs_attestation_service_operator.fs_attestation_service_set_constraints(args.signer_ip_address_constraints, args.expiry_time_constrint, args.risk_score_threshold_constraint)

    @safe_exception
    def fs_attestation_settings(args):
        fs_attestation_service_operator.fs_attestation_settings(args.contact_email_address, args.slack_url)

    # Cli owners
    parser_show_cli_owners = subparsers.add_parser("show_cli_owners")
    parser_show_cli_owners.set_defaults(func=show_cli_owners)

    parser_load_cli_owners_from_words = subparsers.add_parser(
        "load_cli_owners_from_words"
    )
    parser_load_cli_owners_from_words.add_argument("words", type=str, nargs="+")
    parser_load_cli_owners_from_words.set_defaults(func=load_cli_owners_from_words)

    parser_load_cli_owners = subparsers.add_parser("load_cli_owners")
    parser_load_cli_owners.add_argument("keys", type=str, nargs="+")
    parser_load_cli_owners.set_defaults(func=load_cli_owners)

    parser_load_ledger_cli_owners = subparsers.add_parser("load_ledger_cli_owners")
    parser_load_ledger_cli_owners.add_argument(
        "--derivation-path",
        type=str,
        help="Load address for the provided derivation path",
    )
    parser_load_ledger_cli_owners.add_argument(
        "--legacy-accounts",
        action="store_true",
        help="Search for legacy accounts",
    )
    parser_load_ledger_cli_owners.set_defaults(func=load_ledger_cli_owners)

    parser_load_trezor_cli_owners = subparsers.add_parser("load_trezor_cli_owners")
    parser_load_trezor_cli_owners.add_argument(
        "--derivation-path",
        type=str,
        help="Load address for the provided derivation path",
    )
    parser_load_trezor_cli_owners.add_argument(
        "--legacy-accounts",
        action="store_true",
        help="Search for legacy accounts",
    )
    parser_load_trezor_cli_owners.set_defaults(func=load_trezor_cli_owners)

    parser_unload_cli_owners = subparsers.add_parser("unload_cli_owners")
    parser_unload_cli_owners.add_argument(
        "addresses", type=check_ethereum_address, nargs="+"
    )
    parser_unload_cli_owners.set_defaults(func=unload_cli_owners)

    # Change threshold
    parser_change_threshold = subparsers.add_parser("change_threshold")
    parser_change_threshold.add_argument("threshold", type=int)
    parser_change_threshold.set_defaults(func=change_threshold)

    # Approve hash
    parser_approve_hash = subparsers.add_parser("approve_hash")
    parser_approve_hash.add_argument("hash_to_approve", type=check_keccak256_hash)
    parser_approve_hash.add_argument("sender", type=check_ethereum_address)
    parser_approve_hash.set_defaults(func=approve_hash)

    # Sign message
    parser_sign_message = subparsers.add_parser("sign_message")
    group = parser_sign_message.add_mutually_exclusive_group(required=True)
    group.add_argument("--eip191_message", type=str)
    group.add_argument("--eip712_path", type=str)
    parser_sign_message.set_defaults(func=sign_message)

    # Add owner
    parser_add_owner = subparsers.add_parser("add_owner")
    parser_add_owner.add_argument("address", type=check_ethereum_address)
    parser_add_owner.add_argument("--threshold", type=int, default=None)
    parser_add_owner.set_defaults(func=add_owner)

    # Remove owner
    parser_remove_owner = subparsers.add_parser("remove_owner")
    parser_remove_owner.add_argument("address", type=check_ethereum_address)
    parser_remove_owner.add_argument("--threshold", type=int, default=None)
    parser_remove_owner.set_defaults(func=remove_owner)

    # Change FallbackHandler
    parser_change_fallback_handler = subparsers.add_parser("change_fallback_handler")
    parser_change_fallback_handler.add_argument("address", type=check_ethereum_address)
    parser_change_fallback_handler.set_defaults(func=change_fallback_handler)

    # Change Guard
    parser_change_guard = subparsers.add_parser("change_guard")
    parser_change_guard.add_argument("address", type=check_ethereum_address)
    parser_change_guard.set_defaults(func=change_guard)

    # Change MasterCopy
    parser_change_master_copy = subparsers.add_parser("change_master_copy")
    parser_change_master_copy.add_argument("address", type=check_ethereum_address)
    parser_change_master_copy.set_defaults(func=change_master_copy)

    # Update Safe to last version
    parser_update_version = subparsers.add_parser("update")
    parser_update_version.set_defaults(func=update_version)

    # Update non L2 Safe to L2 Safe
    parser_update_version_to_l2 = subparsers.add_parser("update_version_to_l2")
    parser_update_version_to_l2.add_argument(
        "migration_contract", type=check_ethereum_address
    )
    parser_update_version_to_l2.set_defaults(func=update_version_to_l2)

    # Send custom/ether/erc20/erc721
    parser_send_custom = subparsers.add_parser("send_custom")
    parser_send_ether = subparsers.add_parser("send_ether")
    parser_send_erc20 = subparsers.add_parser("send_erc20")
    parser_send_erc721 = subparsers.add_parser("send_erc721")
    parser_send_custom.set_defaults(func=send_custom)
    parser_send_ether.set_defaults(func=send_ether)
    parser_send_erc20.set_defaults(func=send_erc20)
    parser_send_erc721.set_defaults(func=send_erc721)
    # They have some common arguments
    for parser in (
        parser_send_custom,
        parser_send_ether,
        parser_send_erc20,
        parser_send_erc721,
    ):
        parser.add_argument(
            "--safe-nonce",
            type=int,
            help="Use custom safe nonce instead of "
            "the one for last executed SafeTx + 1",
        )

    # To/value is common for send custom and send ether
    for parser in (parser_send_custom, parser_send_ether):
        parser.add_argument("to", type=check_ethereum_address)
        parser.add_argument("value", type=int)

    parser_send_custom.add_argument("data", type=check_hex_str)
    parser_send_custom.add_argument(
        "--delegate", action="store_true", help="Use DELEGATE_CALL. By default use CALL"
    )

    # Send erc20/721 have common arguments
    for parser in (parser_send_erc20, parser_send_erc721):
        parser.add_argument("to", type=check_ethereum_address)
        parser.add_argument("token_address", type=check_ethereum_address)

    parser_send_erc20.add_argument("amount", type=int)
    parser_send_erc721.add_argument("token_id", type=int)

    # Drain only needs receiver account
    parser_drain = subparsers.add_parser("drain")
    parser_drain.set_defaults(func=drain)
    parser_drain.add_argument("to", type=check_ethereum_address)

    # Retrieve threshold, nonce or owners
    parser_get_threshold = subparsers.add_parser("get_threshold")
    parser_get_threshold.set_defaults(func=get_threshold)

    parser_get_nonce = subparsers.add_parser("get_nonce")
    parser_get_nonce.set_defaults(func=get_nonce)

    parser_get_owners = subparsers.add_parser("get_owners")
    parser_get_owners.set_defaults(func=get_owners)

    # Enable and disable modules
    parser_enable_module = subparsers.add_parser("enable_module")
    parser_enable_module.add_argument("address", type=check_ethereum_address)
    parser_enable_module.set_defaults(func=enable_module)

    parser_disable_module = subparsers.add_parser("disable_module")
    parser_disable_module.add_argument("address", type=check_ethereum_address)
    parser_disable_module.set_defaults(func=disable_module)

    # Info and refresh
    parser_info = subparsers.add_parser("info")
    parser_info.set_defaults(func=get_info)

    parser_refresh = subparsers.add_parser("refresh")
    parser_refresh.set_defaults(func=get_refresh)

    # Tx-Service
    # TODO Use subcommands
    parser_tx_service = subparsers.add_parser("balances")
    parser_tx_service.set_defaults(func=get_balances)

    parser_tx_service = subparsers.add_parser("history")
    parser_tx_service.set_defaults(func=get_history)

    parser_tx_service = subparsers.add_parser("sign-tx")
    parser_tx_service.set_defaults(func=sign_tx)
    parser_tx_service.add_argument("safe_tx_hash", type=check_keccak256_hash)

    parser_tx_service = subparsers.add_parser("batch-txs")
    parser_tx_service.set_defaults(func=batch_txs)
    parser_tx_service.add_argument("safe_nonce", type=int)
    parser_tx_service.add_argument(
        "safe_tx_hashes", type=check_keccak256_hash, nargs="+"
    )

    parser_tx_service = subparsers.add_parser("execute-tx")
    parser_tx_service.set_defaults(func=execute_tx)
    parser_tx_service.add_argument("safe_tx_hash", type=check_keccak256_hash)

    # List delegates
    parser_delegates = subparsers.add_parser("get_delegates")
    parser_delegates.set_defaults(func=get_delegates)

    # Add delegate
    parser_add_delegate = subparsers.add_parser("add_delegate")
    parser_add_delegate.set_defaults(func=add_delegate)
    parser_add_delegate.add_argument("address", type=check_ethereum_address)
    parser_add_delegate.add_argument("label", type=str)
    parser_add_delegate.add_argument("signer", type=check_ethereum_address)

    # Remove delegate
    parser_remove_delegate = subparsers.add_parser("remove_delegate")
    parser_remove_delegate.set_defaults(func=remove_delegate)
    parser_remove_delegate.add_argument("address", type=check_ethereum_address)
    parser_remove_delegate.add_argument("signer", type=check_ethereum_address)

    # Remove not executed proposed transaction
    parser_remove_proposed_transaction = subparsers.add_parser(
        "remove_proposed_transaction"
    )
    parser_remove_proposed_transaction.set_defaults(func=remove_proposed_transaction)
    parser_remove_proposed_transaction.add_argument(
        "safe_tx_hash", type=check_keccak256_hash
    )

    # FailSafe Attestation Service Auth
    parser_fs_attestation_service_auth = subparsers.add_parser(
        "fs_attestation_service_auth"
    )
    parser_fs_attestation_service_auth.set_defaults(func=fs_attestation_service_auth)
    parser_fs_attestation_service_auth.add_argument(
        "safe_cli_owner", type=str
    )
    parser_fs_attestation_service_auth.add_argument(
        "safe_cli_owner_private_key", type=str
    )

    # FailSafe Attestation Service Enrollment
    parser_fs_attestation_service_enrolment = subparsers.add_parser(
        "fs_attestation_service_enrolment"
    )
    parser_fs_attestation_service_enrolment.set_defaults(func=fs_attestation_service_enrolment)

    # FailSafe Attestation Service Deploy Guard Contract
    parser_fs_attestation_service_deploy_guard_contract = subparsers.add_parser(
        "fs_attestation_service_deploy_guard_contract"
    )
    parser_fs_attestation_service_deploy_guard_contract.set_defaults(func=fs_attestation_service_deploy_guard_contract)

    # FailSafe Attestation Service Set Constraints
    parser_fs_attestation_service_set_constraints = subparsers.add_parser(
        "fs_attestation_service_set_constraints"
    )
    parser_fs_attestation_service_set_constraints.set_defaults(func=fs_attestation_service_set_constraints)
    parser_fs_attestation_service_set_constraints.add_argument(
        "signer_ip_address_constraints", type=parse_signers_ip_address_constraints, help="Comma separated signer wallet address & IP address constraint: Example - 0x649D4d4cd5E09D02c9717EC51066A39FC843b744:172.21.0.8-172.21.0.30, 0x649D4d4cd5E09D02c9717EC51066A39FC843b745:172.21.1.8-172.21.2.30"
    )
    parser_fs_attestation_service_set_constraints.add_argument(
        "expiry_time_constrint", type=int, help="Maximum duration in days within which all the signers should the sign the multi-sig transaction"
    )
    parser_fs_attestation_service_set_constraints.add_argument(
        "risk_score_threshold_constraint", type=float, help="Maximum risk score threshold for a signer wallet address"
    )

    # FailSafe Attestation Service Settings
    parser_fs_attestation_settings = subparsers.add_parser(
        "fs_attestation_settings"
    )
    parser_fs_attestation_settings.set_defaults(func=fs_attestation_settings)
    parser_fs_attestation_settings.add_argument(
        "contact_email_address", type=str
    )
    parser_fs_attestation_settings.add_argument(
        "slack_url", type=str
    )

    return prompt_parser


from prompt_toolkit import HTML
from prompt_toolkit.completion import CompleteEvent, Completer, Completion
from prompt_toolkit.document import Document

from .safe_completer_constants import (
    meta,
    safe_color_arguments,
    safe_commands,
    safe_commands_arguments,
)


class SafeCompleter(Completer):
    """
    This class will handle auto-completion of known user input commands
    """

    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Completion:
        """
        Provide command completion for function names and param types

        :param document:
        :param complete_event:
        :return:
        """
        word = document.get_word_before_cursor()
        if document.find_previous_word_ending() is None:
            for command in safe_commands:
                # note: force lower() to function as ignore_case.
                if command.startswith(word.lower()):
                    if command in safe_commands_arguments:
                        safe_command = safe_commands_arguments[command]
                        safe_argument_color = safe_color_arguments.get(
                            safe_command, "default"
                        )
                        display = HTML(
                            "<b><ansired> &gt; </ansired>%s</b> <"
                            + safe_argument_color
                            + ">%s</"
                            + safe_argument_color
                            + ">"
                        ) % (command, safe_command)
                        yield Completion(
                            command,
                            start_position=-len(word),
                            display=display,
                            display_meta=meta.get(command),
                        )



class SafeOperatorException(Exception):
    pass


class ExistingOwnerException(SafeOperatorException):
    pass


class NonExistingOwnerException(SafeOperatorException):
    pass


class HashAlreadyApproved(SafeOperatorException):
    pass


class ThresholdLimitException(SafeOperatorException):
    pass


class SameFallbackHandlerException(SafeOperatorException):
    pass


class InvalidFallbackHandlerException(SafeOperatorException):
    pass


class FallbackHandlerNotSupportedException(SafeOperatorException):
    pass


class SameGuardException(SafeOperatorException):
    pass


class InvalidGuardException(SafeOperatorException):
    pass


class GuardNotSupportedException(SafeOperatorException):
    pass


class SameMasterCopyException(SafeOperatorException):
    pass


class SafeAlreadyUpdatedException(SafeOperatorException):
    pass


class SafeVersionNotSupportedException(SafeOperatorException):
    pass


class UpdateAddressesNotValid(SafeOperatorException):
    pass


class SenderRequiredException(SafeOperatorException):
    pass


class AccountNotLoadedException(SafeOperatorException):
    pass


class NotEnoughSignatures(SafeOperatorException):
    pass


class InvalidMasterCopyException(SafeOperatorException):
    pass


class InvalidMigrationContractException(SafeOperatorException):
    pass


class InvalidNonceException(SafeOperatorException):
    pass


class NotEnoughEtherToSend(SafeOperatorException):
    pass


class NotEnoughTokenToSend(SafeOperatorException):
    pass


class SafeServiceNotAvailable(SafeOperatorException):
    pass


class HardwareWalletException(SafeOperatorException):
    pass



from unittest import mock

from eth_account import Account

from gnosis.eth import EthereumClient, EthereumNetwork
from gnosis.safe.tests.safe_test_case import SafeTestCaseMixin

from safe_cli.operators import SafeOperator, SafeOperatorMode, SafeTxServiceOperator


class SafeCliTestCaseMixin(SafeTestCaseMixin):
    def setup_operator(
        self,
        number_owners: int = 1,
        version: str = "1.4.1",
        mode: SafeOperatorMode = SafeOperatorMode.BLOCKCHAIN,
    ) -> SafeOperator:
        assert number_owners >= 1, "Number of owners cannot be less than 1!"
        if version == "1.0.0":
            safe = self.deploy_test_safe_v1_0_0(
                owners=[self.ethereum_test_account.address]
            )
        elif version == "1.1.1":
            safe = self.deploy_test_safe_v1_1_1(
                owners=[self.ethereum_test_account.address]
            )
        elif version == "1.3.0":
            safe = self.deploy_test_safe_v1_3_0(
                owners=[self.ethereum_test_account.address]
            )
        elif version == "1.4.1":
            safe = self.deploy_test_safe_v1_4_1(
                owners=[self.ethereum_test_account.address]
            )
        else:
            raise ValueError(f"{version} not supported")
        if mode == SafeOperatorMode.BLOCKCHAIN:
            safe_operator = SafeOperator(safe.address, self.ethereum_node_url)
        else:
            with mock.patch.object(
                EthereumClient, "get_network", return_value=EthereumNetwork.GOERLI
            ):
                safe_operator = SafeTxServiceOperator(
                    safe.address, self.ethereum_node_url
                )
        safe_operator.load_cli_owners([self.ethereum_test_account.key.hex()])
        for _ in range(number_owners - 1):
            account = Account.create()
            safe_operator.add_owner(account.address)
            safe_operator.load_cli_owners([account.key.hex()])
        return safe_operator



import setuptools

from safe_cli.version import version

with open("README.md", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="fs_attestation_safe_cli",
    version=version,
    author="Mohan",
    author_email="mohan@ethlas.com",
    description="Modified Command Line Interface for FailSafe Attestation Service to administrate Gnosis Safe Wallets and FailSafe Attestation Guard Contracts",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gnosis/safe-cli",
    download_url="https://github.com/gnosis/safe-cli/releases",
    license="MIT",
    test_suite="tests",
    install_requires=[
        "art>=6",
        "colorama>=0.4",
        "prompt_toolkit>=3",
        "pygments>=2",
        "requests>=2",
        "fs-attestation-safe-eth-py==1.0.2",
        "tabulate>=0.8",
    ],
    extras_require={"ledger": ["ledgereth==0.9.1"], "trezor": ["trezor==0.13.8"]},
    packages=setuptools.find_packages(),
    entry_points={
        "console_scripts": [
            "fs-attestation-safe-cli=safe_cli.main:main",
            "safe-creator=safe_cli.safe_creator:main",
        ],
    },
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
)



import dataclasses
import json
import os
from functools import cached_property, wraps
from typing import List, Optional, Sequence, Set, Tuple

from ens import ENS
from eth_account import Account
from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress
from eth_utils import ValidationError
from hexbytes import HexBytes
from packaging import version as semantic_version
from prompt_toolkit import HTML, print_formatted_text
from web3 import Web3
from web3.contract import Contract
from web3.exceptions import BadFunctionCallOutput

from gnosis.eth import (
    EthereumClient,
    EthereumNetwork,
    EthereumNetworkNotSupported,
    TxSpeed,
)
from gnosis.eth.clients import EtherscanClient, EtherscanClientConfigurationProblem
from gnosis.eth.constants import NULL_ADDRESS, SENTINEL_ADDRESS
from gnosis.eth.contracts import (
    get_erc20_contract,
    get_erc721_contract,
    get_safe_V1_1_1_contract,
    get_sign_message_lib_contract,
)
from gnosis.eth.eip712 import eip712_encode
from gnosis.eth.utils import get_empty_tx_params
from gnosis.safe import InvalidInternalTx, Safe, SafeOperationEnum, SafeTx
from gnosis.safe.api import TransactionServiceApi
from gnosis.safe.multi_send import MultiSend, MultiSendOperation, MultiSendTx
from gnosis.safe.safe_deployments import safe_deployments

from safe_cli.ethereum_hd_wallet import get_account_from_words
from safe_cli.operators.exceptions import (
    AccountNotLoadedException
)
from safe_cli.safe_addresses import (
    get_default_fallback_handler_address,
    get_last_sign_message_lib_address,
    get_safe_contract_address,
    get_safe_l2_contract_address,
)
from safe_cli.utils import choose_option_from_list, get_erc_20_list, yes_or_no_question

from ..contracts import safe_to_l2_migration
from .hw_wallets.hw_wallet import HwWallet
from .hw_wallets.hw_wallet_manager import HwWalletType, get_hw_wallet_manager

import requests

from safe_cli.operators.fs_attestation_auth import FailSafeAttestationServiceAuth

@dataclasses.dataclass
class SafeCliInfo:
    address: str
    nonce: int
    threshold: int
    owners: List[str]
    master_copy: str
    modules: List[str]
    fallback_handler: str
    guard: str
    balance_ether: int
    version: str

    def __str__(self):
        return (
            f"safe-version={self.version} nonce={self.nonce} threshold={self.threshold} owners={self.owners} "
            f"master-copy={self.master_copy} fallback-hander={self.fallback_handler} "
            f"modules={self.modules} balance-ether={self.balance_ether:.4f}"
        )

class FailSafeAttestationServiceOperator:
    address: ChecksumAddress
    node_url: str
    ethereum_client: EthereumClient
    ens: ENS
    network: EthereumNetwork
    etherscan: Optional[EtherscanClient]
    safe_tx_service: Optional[TransactionServiceApi]
    safe: Safe
    safe_contract: Contract
    safe_contract_1_1_0: Contract
    accounts: Set[LocalAccount] = set()
    default_sender: Optional[LocalAccount]
    executed_transactions: List[str]
    _safe_cli_info: Optional[SafeCliInfo]
    require_all_signatures: bool

    def __init__(self, address: ChecksumAddress, node_url: str, safe_tx_service_base_url: str, fs_attestation_service_base_url: str):
        self.address = address
        self.node_url = node_url
        self.safe_tx_service_base_url = safe_tx_service_base_url
        self.fs_attestation_service_base_url = fs_attestation_service_base_url
        self.ethereum_client = EthereumClient(self.node_url)
        self.ens = ENS.from_web3(self.ethereum_client.w3)
        self.network: EthereumNetwork = self.ethereum_client.get_network()
        try:
            self.etherscan = EtherscanClient(self.network)
        except EtherscanClientConfigurationProblem:
            self.etherscan = None

        try:
            self.safe_tx_service = TransactionServiceApi.from_custom_safe_tx_service_base_url(
                self.ethereum_client,
                self.safe_tx_service_base_url
            )
        except EthereumNetworkNotSupported:
            self.safe_tx_service = None

        self.safe = Safe(address, self.ethereum_client)
        self.safe_contract = self.safe.contract
        self.safe_contract_1_1_0 = get_safe_V1_1_1_contract(
            self.ethereum_client.w3, address=self.address
        )
        self.accounts: Set[LocalAccount] = set()
        self.default_sender: Optional[LocalAccount] = None
        self.executed_transactions: List[str] = []
        self._safe_cli_info: Optional[SafeCliInfo] = None  # Cache for SafeCliInfo
        self.require_all_signatures = (
            True  # Require all signatures to be present to send a tx
        )
        self.hw_wallet_manager = get_hw_wallet_manager()

    def fs_attestation_user_login(self, safe_cli_owner: str, safe_cli_owner_private_key: str) -> str:
        try:
            # Perform Authentication with FailSafe Attestation Service and get the access token
            print_formatted_text(
                HTML("<ansigreen>Performing FailSafe Attestation Service Authentication. Please wait...</ansigreen>")
            )
            self.fs_attestation_service_auth = FailSafeAttestationServiceAuth(self.fs_attestation_service_base_url, safe_cli_owner, safe_cli_owner_private_key)
            self.access_token = self.fs_attestation_service_auth.fs_attestation_user_login()
            response_text = HTML("<ansigreen>FailSafe Attestation Service Authentication flow completed successfully</ansigreen>")
            print_formatted_text(
                response_text
            )
        except Exception as error:
            response_text = HTML("<ansired>Some error occurred when performing FailSafe Attestation Service Authentication flow</ansired>")
            print_formatted_text(
                response_text
            )

    def fs_attestation_service_enrolment(self) -> str:
        try:
            enrolment_url = "{}/{}".format(self.fs_attestation_service_base_url, "attestation-service/enrollment")
            headers = {'Authorization': 'Bearer {}'.format(self.access_token)}
            data = {'multisig_wallet_address': self.address, 'multisig_wallet_provider': 'gnosis_safe', 'chain_id': self.network.value}
            # Make the POST request
            response = requests.post(enrolment_url, json=data, headers=headers)

            # Check if the request was successful (status code 200)
            if response.status_code == 200:
                # Print the response content
                response_text = HTML("<ansigreen>{response_text}</ansigreen>").format(response_text=response.text)
                print_formatted_text(
                    response_text
                )
                return response.text
            else:
                # Print an error message if the request was not successful
                response_text = HTML("<ansired>Status code {response_code}: {response_text}</ansired>").format(response_code=response.status_code, response_text=response.text)
                print_formatted_text(
                    response_text
                )
                return ""
        except Exception as error:
            response_text = HTML("<ansired>Some error occurred when performing FailSafe Attestation Service Enrolment</ansired>")
            print_formatted_text(
                response_text
            )

    def fs_attestation_service_deploy_guard_contract(self) -> str:
        try:
            deploy_guard_contract_url = "{}/{}".format(self.fs_attestation_service_base_url, "attestation-service/safe-guard-contract-deployment")
            headers = {'Authorization': 'Bearer {}'.format(self.access_token)}
            data = {'multisig_wallet_address': self.address, 'chain_id': self.network.value}
            # Make the POST request
            response = requests.post(deploy_guard_contract_url, json=data, headers=headers)

            # Check if the request was successful (status code 200)
            if response.status_code == 200:
                # Print the response content
                response_text = HTML("<ansigreen>{response_text}</ansigreen>").format(response_text=response.text)
                print_formatted_text(
                    response_text
                )
                return response.text
            else:
                # Print an error message if the request was not successful
                response_text = HTML("<ansired>Status code {response_code}: {response_text}</ansired>").format(response_code=response.status_code, response_text=response.text)
                print_formatted_text(
                    response_text
                )
                return ""
        except Exception as error:
            response_text = HTML("<ansired>Some error occurred when deploying Guard Contract for the Safe Multi-Sig Wallet</ansired>")
            print_formatted_text(
                response_text
            )

    def fs_attestation_service_set_constraints(self, signer_ip_address_constraints: list, expiry_time_constrint: str, risk_score_threshold_constraint: float) -> str:
        try:
            set_constraints_url = "{}/{}".format(self.fs_attestation_service_base_url, "attestation-service/constraints")
            headers = {'Authorization': 'Bearer {}'.format(self.access_token)}
            data = {'multisig_wallet_address': self.address, 'chain_id': self.network.value, 'constraints':{
                'signers': signer_ip_address_constraints,
                'expiry_time': {"unit": "d", "value": str(expiry_time_constrint)},
                'risk_score': risk_score_threshold_constraint
            }}

            # Make the POST request
            response = requests.post(set_constraints_url, json=data, headers=headers)

            # Check if the request was successful (status code 200)
            if response.status_code == 200:
                # Print the response content
                response_text = HTML("<ansigreen>{response_text}</ansigreen>").format(response_text=response.text)
                print_formatted_text(
                    response_text
                )
                return response.text
            else:
                # Print an error message if the request was not successful
                response_text = HTML("<ansired>Status code {response_code}: {response_text}</ansired>").format(response_code=response.status_code, response_text=response.text)
                print_formatted_text(
                    response_text
                )
                return ""
        except Exception as error:
            response_text = HTML("<ansired>Some error occurred when setting FailSafe Attestation constraints for the Safe Multi-Sig Wallet</ansired>")
            print_formatted_text(
                response_text
            )

    def fs_attestation_settings(self, contact_email_address: str, slack_url: str) -> str:
        try:
            set_constraints_url = "{}/{}".format(self.fs_attestation_service_base_url, "attestation-service/settings")
            headers = {'Authorization': 'Bearer {}'.format(self.access_token)}
            data = {'multisig_wallet_address': self.address, 'chain_id': self.network.value, 'contact_email_address': contact_email_address, 'slack_url': slack_url}

            # Make the POST request
            response = requests.post(set_constraints_url, json=data, headers=headers)

            # Check if the request was successful (status code 200)
            if response.status_code == 200:
                # Print the response content
                response_text = HTML("<ansigreen>{response_text}</ansigreen>").format(response_text=response.text)
                print_formatted_text(
                    response_text
                )
                return response.text
            else:
                # Print an error message if the request was not successful
                response_text = HTML("<ansired>Status code {response_code}: {response_text}</ansired>").format(response_code=response.status_code, response_text=response.text)
                print_formatted_text(
                    response_text
                )
                return ""
        except Exception as error:
            response_text = HTML("<ansired>Some error occurred when updating the FailSafe Attestation Service settings for the Safe Multi-Sig Wallet</ansired>")
            print_formatted_text(
                response_text
            )



import unittest

from eth_account import Account

from gnosis.eth import EthereumClient
from gnosis.eth.tests.utils import just_test_if_mainnet_node

from safe_cli.safe_addresses import (
    _get_valid_contract,
    get_default_fallback_handler_address,
    get_last_multisend_address,
    get_last_multisend_call_only_address,
    get_proxy_factory_address,
    get_safe_contract_address,
    get_safe_l2_contract_address,
)

from .safe_cli_test_case_mixin import SafeCliTestCaseMixin


class TestSafeAddresses(SafeCliTestCaseMixin, unittest.TestCase):
    def test_get_valid_contract(self):
        addresses = [
            Account.create().address,
            Account.create().address,
            self.safe_contract_V1_4_1.address,
            Account.create().address,
        ]
        expected_address = self.safe_contract_V1_4_1.address
        self.assertEqual(
            _get_valid_contract(self.ethereum_client, addresses), expected_address
        )

        with self.assertRaisesRegex(
            ValueError,
            f"Network {self.ethereum_client.get_network().name} is not supported",
        ):
            _get_valid_contract(self.ethereum_client, addresses[:1])

    def test_get_addresses(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)

        self.assertEqual(
            get_safe_contract_address(ethereum_client),
            "0x41675C099F32341bf84BFc5382aF534df5C7461a",
        )
        self.assertEqual(
            get_safe_l2_contract_address(ethereum_client),
            "0x29fcB43b46531BcA003ddC8FCB67FFE91900C762",
        )
        self.assertEqual(
            get_default_fallback_handler_address(ethereum_client),
            "0xfd0732Dc9E303f09fCEf3a7388Ad10A83459Ec99",
        )
        self.assertEqual(
            get_proxy_factory_address(ethereum_client),
            "0x4e1DCf7AD4e460CfD30791CCC4F9c8a4f820ec67",
        )
        self.assertEqual(
            get_last_multisend_address(ethereum_client),
            "0x38869bf66a61cF6bDB996A6aE40D5853Fd43B526",
        )
        self.assertEqual(
            get_last_multisend_call_only_address(ethereum_client),
            "0x40A2aCCbd92BCA938b02010E17A5b8929b49130D",
        )


if __name__ == "__main__":
    unittest.main()



from enum import Enum
from functools import lru_cache
from typing import Dict, List, Optional, Set, Tuple

from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from prompt_toolkit import HTML, print_formatted_text
from web3.types import TxParams, Wei

from gnosis.eth import TxSpeed
from gnosis.eth.eip712 import eip712_encode, eip712_encode_hash
from gnosis.safe import SafeTx
from gnosis.safe.safe_signature import SafeSignature, SafeSignatureEOA

from .hw_wallet import HwWallet


class HwWalletType(Enum):
    TREZOR = 0
    LEDGER = 1


@lru_cache(maxsize=None)
def get_hw_wallet_manager():
    return HwWalletManager()


class HwWalletManager:
    def __init__(self):
        self.wallets: Set[HwWallet] = set()
        self.supported_hw_wallet_types: Dict[str, HwWallet] = {}
        self.sender: Optional[HwWallet] = None
        try:
            from .ledger_wallet import LedgerWallet

            self.supported_hw_wallet_types[HwWalletType.LEDGER] = LedgerWallet
        except ImportError:
            pass

        try:
            from .trezor_wallet import TrezorWallet

            self.supported_hw_wallet_types[HwWalletType.TREZOR] = TrezorWallet
        except ImportError:
            pass

    def is_supported_hw_wallet(self, hw_wallet_type: HwWalletType) -> bool:
        return hw_wallet_type in self.supported_hw_wallet_types

    def get_hw_wallet(self, hw_wallet_type: HwWalletType) -> Optional[HwWallet]:
        if hw_wallet_type in self.supported_hw_wallet_types:
            return self.supported_hw_wallet_types[hw_wallet_type]

    def get_accounts(
        self,
        hw_wallet_type: HwWalletType,
        template_derivation_path: str,
        number_accounts: Optional[int] = 5,
    ) -> List[Tuple[ChecksumAddress, str]]:
        """

        :param hw_wallet: Trezor or Ledger
        :param template_derivation_path: formatted string to indicate which path iterate
        :param number_accounts:  number of accounts requested to ledger
        :return: a list of tuples with address and derivation path
        """
        accounts = []
        hw_wallet = self.get_hw_wallet(hw_wallet_type)
        for i in range(number_accounts):
            path_string = template_derivation_path.format(i=i)
            accounts.append((hw_wallet(path_string).address, path_string))
        return accounts

    def add_account(
        self, hw_wallet_type: HwWalletType, derivation_path: str
    ) -> ChecksumAddress:
        """
        Add an account to ledger manager set and return the added address

        :param derivation_path:
        :return:
        """

        hw_wallet = self.get_hw_wallet(hw_wallet_type)

        wallet = hw_wallet(derivation_path)
        self.wallets.add(wallet)
        return wallet.address

    def set_sender(self, hw_wallet_type: HwWalletType, derivation_path: str):
        """
        Set a harware wallet as a sender to enable execute transaction from it.
        :param hw_wallet_type:
        :param derivation_path:
        :return:
        """
        hw_wallet = self.get_hw_wallet(hw_wallet_type)
        self.sender = hw_wallet(derivation_path)

    def delete_accounts(self, addresses: List[ChecksumAddress]) -> Set:
        """
        Remove ledger accounts from address

        :param accounts:
        :return: list with the delete accounts
        """
        accounts_to_remove = set()
        for address in addresses:
            for account in self.wallets:
                if account.address == address:
                    if self.sender and self.sender.address == address:
                        self.sender = None
                    accounts_to_remove.add(account)
        self.wallets = self.wallets.difference(accounts_to_remove)
        return accounts_to_remove

    def sign_eip712(self, eip712_message: Dict, wallets: List[HwWallet]) -> HexBytes:
        """
        Sign an EIP712 message

        :param eip712_message:
        :param wallets:
        :return: Appended sorted signatures for all the provided wallets
        """
        _, domain_hash, message_hash = eip712_encode(eip712_message)
        eip712_message_hash = eip712_encode_hash(eip712_message)
        safe_signatures: List[SafeSignature] = []
        for wallet in wallets:
            print_formatted_text(
                HTML(
                    f"<ansired>Make sure before signing in your {wallet} that the domain_hash and message_hash are both correct</ansired>"
                )
            )
            print_formatted_text(HTML(f"Domain_hash: <b>{domain_hash.hex()}</b>"))
            print_formatted_text(HTML(f"Message_hash: <b>{message_hash.hex()}</b>"))
            signature = wallet.sign_typed_hash(domain_hash, message_hash)
            safe_signatures.append(SafeSignatureEOA(signature, eip712_message_hash))

        return SafeSignature.export_signatures(safe_signatures)

    def sign_safe_tx(self, safe_tx: SafeTx, wallets: List[HwWallet]) -> SafeTx:
        """
        Sign a safe transaction with the provided hardware wallets

        :param safe_tx:
        :param wallets:
        :return: SafeTx with signature.
        """
        signatures = self.sign_eip712(safe_tx.eip712_structured_data, wallets)
        safe_tx.signatures = signatures
        return safe_tx

    def execute_safe_tx(
        self,
        safe_tx: SafeTx,
        tx_gas: Optional[int] = None,
        tx_gas_price: Optional[int] = None,
        tx_nonce: Optional[int] = None,
        eip1559_speed: Optional[TxSpeed] = None,
    ) -> Tuple[HexBytes, TxParams]:
        """
        Send multisig tx to the Safe

        :param safe_tx: Safe transaction to sign
        :param tx_gas: Gas for the external tx. If not, `(safe_tx_gas + base_gas) * 2` will be used
        :param tx_gas_price: Gas price of the external tx. If not, `gas_price` will be used
        :param tx_nonce: Force nonce for `tx_sender`
        :param eip1559_speed: If provided, use EIP1559 transaction
        :return: Tuple(tx_hash, tx)
        """

        if eip1559_speed and safe_tx.ethereum_client.is_eip1559_supported():
            tx_parameters = safe_tx.ethereum_client.set_eip1559_fees(
                {
                    "from": self.sender.address,
                },
                tx_speed=eip1559_speed,
            )
        else:
            tx_parameters = {
                "from": self.sender.address,
                "gasPrice": tx_gas_price or safe_tx.w3.eth.gas_price,
            }

        if tx_gas:
            tx_parameters["gas"] = tx_gas

        if tx_nonce is not None:
            tx_parameters["nonce"] = tx_nonce
        else:
            tx_parameters["nonce"] = safe_tx.ethereum_client.get_nonce_for_account(
                self.sender.address, block_identifier="latest"
            )
        safe_tx.tx = safe_tx.w3_tx.build_transaction(tx_parameters)
        safe_tx.tx["gas"] = Wei(
            tx_gas or (max(safe_tx.tx["gas"] + 75000, safe_tx.recommended_gas()))
        )
        signed_raw_transaction = self.sender.get_signed_raw_transaction(
            safe_tx.tx, safe_tx.ethereum_client.get_chain_id()
        )  # sign with ledger
        safe_tx.tx_hash = safe_tx.ethereum_client.w3.eth.send_raw_transaction(
            signed_raw_transaction
        )
        # Set signatures empty after executing the tx. `Nonce` is increased even if it fails,
        # so signatures are not valid anymore
        safe_tx.signatures = b""
        return safe_tx.tx_hash, safe_tx.tx



import argparse
import unittest
from unittest import mock
from unittest.mock import MagicMock

from eth_account import Account
from prompt_toolkit import HTML

from gnosis.eth.constants import NULL_ADDRESS
from gnosis.eth.ethereum_client import EthereumClient
from gnosis.safe import Safe
from gnosis.safe.api import TransactionServiceApi
from gnosis.safe.safe import SafeInfo

from safe_cli.main import SafeCli, build_safe_cli
from safe_cli.operators import SafeOperator

from .safe_cli_test_case_mixin import SafeCliTestCaseMixin


class SafeCliEntrypointTestCase(SafeCliTestCaseMixin, unittest.TestCase):
    random_safe_address = Account.create().address

    @mock.patch("argparse.ArgumentParser.parse_args")
    def build_test_safe_cli(self, mock_parse_args: MagicMock):
        mock_parse_args.return_value = argparse.Namespace(
            address=self.random_safe_address,
            node_url=self.ethereum_node_url,
            history=True,
            get_safes_from_owner=False,
        )
        return build_safe_cli()

    @mock.patch("argparse.ArgumentParser.parse_args")
    def build_test_safe_cli_for_owner(self, mock_parse_args: MagicMock):
        mock_parse_args.return_value = argparse.Namespace(
            address=self.random_safe_address,
            node_url=self.ethereum_node_url,
            history=True,
            get_safes_from_owner=True,
        )
        return build_safe_cli()

    @mock.patch.object(Safe, "retrieve_all_info")
    def test_build_safe_cli(self, retrieve_all_info_mock: MagicMock):
        retrieve_all_info_mock.return_value = SafeInfo(
            self.random_safe_address,
            "0xfd0732Dc9E303f09fCEf3a7388Ad10A83459Ec99",
            NULL_ADDRESS,
            "0x29fcB43b46531BcA003ddC8FCB67FFE91900C762",
            [],
            0,
            [Account.create().address],
            1,
            "1.4.1",
        )

        safe_cli = self.build_test_safe_cli()
        with mock.patch.object(SafeOperator, "is_version_updated", return_value=True):
            self.assertIsNone(safe_cli.print_startup_info())
        self.assertIsInstance(safe_cli.get_prompt_text(), HTML)
        self.assertIsInstance(safe_cli.get_bottom_toolbar(), HTML)

    @mock.patch.object(EthereumClient, "get_chain_id", return_value=5)
    @mock.patch.object(TransactionServiceApi, "get_safes_for_owner")
    @mock.patch.object(Safe, "retrieve_all_info")
    def test_build_safe_cli_for_owner(
        self,
        retrieve_all_info_mock: MagicMock,
        get_safes_for_owner_mock: MagicMock,
        get_chain_id_mock: MagicMock,
    ):
        retrieve_all_info_mock.return_value = SafeInfo(
            self.random_safe_address,
            "0xfd0732Dc9E303f09fCEf3a7388Ad10A83459Ec99",
            NULL_ADDRESS,
            "0x29fcB43b46531BcA003ddC8FCB67FFE91900C762",
            [],
            0,
            [Account.create().address],
            1,
            "1.4.1",
        )
        get_safes_for_owner_mock.return_value = []
        with self.assertRaises(ValueError):
            self.build_test_safe_cli_for_owner()
        get_safes_for_owner_mock.return_value = [self.random_safe_address]
        safe_cli = self.build_test_safe_cli_for_owner()
        self.assertIsNotNone(safe_cli)
        with mock.patch.object(SafeOperator, "is_version_updated", return_value=True):
            self.assertIsNone(safe_cli.print_startup_info())
        self.assertIsInstance(safe_cli.get_prompt_text(), HTML)
        self.assertIsInstance(safe_cli.get_bottom_toolbar(), HTML)

    def test_parse_operator_mode(self):
        safe_cli = self.build_test_safe_cli()
        self.assertIsNone(safe_cli.parse_operator_mode("tx-service"))
        self.assertIsInstance(safe_cli.parse_operator_mode("blockchain"), SafeOperator)

    @mock.patch.object(SafeCli, "get_command", side_effect=EOFError)
    def test_loop(self, mock_parse_args: MagicMock):
        safe_cli = self.build_test_safe_cli()
        safe_cli.loop()


if __name__ == "__main__":
    unittest.main()



import unittest

from pygments.token import Token

from safe_cli.safe_lexer import SafeLexer


class TestSafeLexer(unittest.TestCase):
    def test_get_tokens_unprocessed(self):
        safe_lexer = SafeLexer()

        self.assertEqual(
            next(safe_lexer.get_tokens_unprocessed("refresh")),
            (0, Token.Name.Builtin, "refresh"),
        )
        self.assertEqual(
            next(safe_lexer.get_tokens_unprocessed("not-supported-command")),
            (0, Token.Text, "not-supported-command"),
        )


if __name__ == "__main__":
    unittest.main()



balances_mock = [
    {"tokenAddress": None, "token": None, "balance": "171000000000000000"},
    {
        "tokenAddress": "0x70cBa46d2e933030E2f274AE58c951C800548AeF",
        "token": {
            "name": "Basic Attention Token",
            "symbol": "BAT",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x70cBa46d2e933030E2f274AE58c951C800548AeF.png",
        },
        "balance": "1000000000000000000",
    },
    {
        "tokenAddress": "0x4644A362B7041aff2CF792fD61671b9F75234C27",
        "token": {
            "name": " best tron wallet | tronscan | tronlink | bttc",
            "symbol": "tronAddress.org",
            "decimals": 0,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x4644A362B7041aff2CF792fD61671b9F75234C27.png",
        },
        "balance": "8888",
    },
    {
        "tokenAddress": "0x91056D4A53E1faa1A84306D4deAEc71085394bC8",
        "token": {
            "name": "CoW Protocol Token",
            "symbol": "COW",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x91056D4A53E1faa1A84306D4deAEc71085394bC8.png",
        },
        "balance": "7455447592176417847",
    },
    {
        "tokenAddress": "0x3430d04E42a722c5Ae52C5Bffbf1F230C2677600",
        "token": {
            "name": "CoW Protocol Token",
            "symbol": "COW",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x3430d04E42a722c5Ae52C5Bffbf1F230C2677600.png",
        },
        "balance": "8045157924038386764",
    },
    {
        "tokenAddress": "0xdc31Ee1784292379Fbb2964b3B9C4124D8F89C60",
        "token": {
            "name": "Dai",
            "symbol": "DAI",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0xdc31Ee1784292379Fbb2964b3B9C4124D8F89C60.png",
        },
        "balance": "1000000000000000000000",
    },
    {
        "tokenAddress": "0x64685e9733d57Ccc3870ceD50ed9B1f576Fd61B1",
        "token": {
            "name": "Elden Ring Token",
            "symbol": "ERT",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x64685e9733d57Ccc3870ceD50ed9B1f576Fd61B1.png",
        },
        "balance": "10000000000000000000000",
    },
    {
        "tokenAddress": "0xD14DD51BE76e6dC28756a442dF786845e03a1D60",
        "token": {
            "name": "ethAddress.io",
            "symbol": "ethAddress.io",
            "decimals": 0,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0xD14DD51BE76e6dC28756a442dF786845e03a1D60.png",
        },
        "balance": "888888",
    },
    {
        "tokenAddress": "0x4c6FFFda6b732b495c974Fb284e9bB38b9f061C8",
        "token": {
            "name": "GiantETHLP",
            "symbol": "gETH",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x4c6FFFda6b732b495c974Fb284e9bB38b9f061C8.png",
        },
        "balance": "1000000000000000",
    },
    {
        "tokenAddress": "0x02ABBDbAaa7b1BB64B5c878f7ac17f8DDa169532",
        "token": {
            "name": "Gnosis Token",
            "symbol": "GNO",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x02ABBDbAaa7b1BB64B5c878f7ac17f8DDa169532.png",
        },
        "balance": "400000000000000000",
    },
    {
        "tokenAddress": "0x1643E812aE58766192Cf7D2Cf9567dF2C37e9B7F",
        "token": {
            "name": "Liquid staked Ether 2.0",
            "symbol": "stETH",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x1643E812aE58766192Cf7D2Cf9567dF2C37e9B7F.png",
        },
        "balance": "30196852081566501",
    },
    {
        "tokenAddress": "0x54FA517F05e11Ffa87f4b22AE87d91Cec0C2D7E1",
        "token": {
            "name": "RareTron.io",
            "symbol": "RareTron.io",
            "decimals": 0,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x54FA517F05e11Ffa87f4b22AE87d91Cec0C2D7E1.png",
        },
        "balance": "66666",
    },
    {
        "tokenAddress": "0x61fD3b6d656F39395e32f46E2050953376c3f5Ff",
        "token": {
            "name": "Safe Token",
            "symbol": "SAFE",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x61fD3b6d656F39395e32f46E2050953376c3f5Ff.png",
        },
        "balance": "10362070000000000000000",
    },
    {
        "tokenAddress": "0x5943F705aBb6834Cad767e6E4bB258Bc48D9C947",
        "token": {
            "name": "Super ETH",
            "symbol": "ETHx",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x5943F705aBb6834Cad767e6E4bB258Bc48D9C947.png",
        },
        "balance": "5000000000000000",
    },
    {
        "tokenAddress": "0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984",
        "token": {
            "name": "Uniswap",
            "symbol": "UNI",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984.png",
        },
        "balance": "21295518536254652",
    },
    {
        "tokenAddress": "0xD87Ba7A50B2E7E660f678A895E4B72E7CB4CCd9C",
        "token": {
            "name": "USD Coin",
            "symbol": "USDC",
            "decimals": 6,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0xD87Ba7A50B2E7E660f678A895E4B72E7CB4CCd9C.png",
        },
        "balance": "10000000",
    },
    {
        "tokenAddress": "0xe0C9275E44Ea80eF17579d33c55136b7DA269aEb",
        "token": {
            "name": "USD Coin",
            "symbol": "USDC",
            "decimals": 6,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0xe0C9275E44Ea80eF17579d33c55136b7DA269aEb.png",
        },
        "balance": "925089",
    },
    {
        "tokenAddress": "0x38d9639da06f7a726cAf40c2E1010ac88517A085",
        "token": {
            "name": "VanityTron.io",
            "symbol": "VanityTron.io",
            "decimals": 0,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x38d9639da06f7a726cAf40c2E1010ac88517A085.png",
        },
        "balance": "6666666",
    },
    {
        "tokenAddress": "0x1B809925ba90c541d895D19f0b7D70eE281a987F",
        "token": {
            "name": "VanityTRX.org",
            "symbol": "VanityTRX.org",
            "decimals": 0,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0x1B809925ba90c541d895D19f0b7D70eE281a987F.png",
        },
        "balance": "888888",
    },
    {
        "tokenAddress": "0xdFCeA9088c8A88A76FF74892C1457C17dfeef9C1",
        "token": {
            "name": "Wrapped Ether",
            "symbol": "WETH",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0xdFCeA9088c8A88A76FF74892C1457C17dfeef9C1.png",
        },
        "balance": "499812357969727",
    },
    {
        "tokenAddress": "0xB4FBF271143F4FBf7B91A5ded31805e42b2208d6",
        "token": {
            "name": "Wrapped Ether",
            "symbol": "WETH",
            "decimals": 18,
            "logoUri": "https://safe-transaction-assets.staging.5afe.dev/tokens/logos/0xB4FBF271143F4FBf7B91A5ded31805e42b2208d6.png",
        },
        "balance": "5600000000000000",
    },
]
